from typing import TYPE_CHECKING, Any, Callable, Coroutine

from fastapi import Request as BaseRequest
from fastapi.routing import APIRoute
from starlette.responses import Response

if TYPE_CHECKING:
    from .registry import AppRegistry  # coverage: ignore


class FastlifeRequest(BaseRequest):
    def __init__(self, registry: "AppRegistry", request: BaseRequest) -> None:
        super().__init__(request.scope, request.receive)
        self.registry = registry


class FastlifeRoute(APIRoute):
    registry: "AppRegistry" = None  # type: ignore

    def get_route_handler(  # type: ignore
        self,
    ) -> Callable[[FastlifeRequest], Coroutine[Any, Any, Response]]:
        orig_route_handler = super().get_route_handler()

        async def route_handler(request: BaseRequest) -> FastlifeRequest:
            req = FastlifeRequest(self.registry, request)
            return await orig_route_handler(req)  # type: ignore

        return route_handler  # type: ignore



from typing import TYPE_CHECKING, Any, Mapping, MutableMapping, Optional, Sequence, Type

from fastapi import Request
from jinjax.catalog import Catalog
from markupsafe import Markup
from pydantic.fields import FieldInfo

from fastlife.request.model_result import ModelResult
from fastlife.templating.renderer.widgets.factory import WidgetFactory

if TYPE_CHECKING:
    from fastlife.configurator.settings import Settings  # coverage: ignore

from fastlife.shared_utils.resolver import resolve, resolve_path

from .abstract import AbstractTemplateRenderer, AbstractTemplateRendererFactory


def build_searchpath(template_search_path: str) -> Sequence[str]:
    searchpath: list[str] = []
    paths = template_search_path.split(",")

    for path in paths:
        if ":" in path:
            searchpath.append(resolve_path(path))
        else:
            searchpath.append(path)
    return searchpath


class JinjaxRenderer(AbstractTemplateRenderer):
    def __init__(
        self,
        catalog: Catalog,
        request: Request,
        csrf_token_name: str,
        form_data_model_prefix: str,
        route_prefix: str,
    ):
        self.route_prefix = route_prefix
        self.catalog = catalog
        self.request = request
        self.csrf_token_name = csrf_token_name
        self.form_data_model_prefix = form_data_model_prefix
        self.globals: MutableMapping[str, Any] = {}

    def build_globals(self) -> Mapping[str, Any]:
        return {
            "request": self.request,
            "csrf_token": {
                "name": self.csrf_token_name,
                "value": self.request.scope.get(self.csrf_token_name, ""),
            },
            "pydantic_form": self.pydantic_form,
            **self.globals,
        }

    def render_template(
        self,
        template: str,
        *,
        globals: Optional[Mapping[str, Any]] = None,
        **params: Any,
    ) -> str:
        if globals:
            self.globals.update(globals)
        return self.catalog.render(  # type: ignore
            template, __globals=self.build_globals(), **params
        )

    def pydantic_form(
        self, model: ModelResult[Any], *, token: Optional[str] = None
    ) -> Markup:
        return WidgetFactory(self, token).get_markup(model)

    def pydantic_form_field(
        self,
        model: Type[Any],
        *,
        name: str | None,
        token: str | None,
        removable: bool,
        field: FieldInfo | None,
    ) -> Markup:
        return (
            WidgetFactory(self, token)
            .get_widget(
                model,
                form_data={},
                form_errors={},
                prefix=(name or self.form_data_model_prefix),
                removable=removable,
                field=field,
            )
            .to_html(self)
        )


class JinjaxTemplateRenderer(AbstractTemplateRendererFactory):
    """
    The default template renderer factory. Based on JinjaX.

    :param settings: setting used to configure jinjax.
    """

    route_prefix: str
    """Used to prefix url to fetch fast life widgets."""

    def __init__(self, settings: "Settings") -> None:
        self.route_prefix = settings.fastlife_route_prefix
        self.form_data_model_prefix = settings.form_data_model_prefix
        self.csrf_token_name = settings.csrf_token_name
        globals = resolve(settings.jinjax_global_catalog_class)().model_dump()

        self.catalog = Catalog(
            use_cache=settings.jinjax_use_cache,
            auto_reload=settings.jinjax_auto_reload,
            globals=globals,
        )
        for path in build_searchpath(settings.template_search_path):
            self.catalog.add_folder(path)

    def __call__(self, request: Request) -> AbstractTemplateRenderer:
        """Build the renderer to render request for template."""
        return JinjaxRenderer(
            self.catalog,
            request,
            self.csrf_token_name,
            self.form_data_model_prefix,
            self.route_prefix,
        )



"""
Prevents CSRF attack using cookie and html hidden field comparaison.

Fast life did not reinvent the wheel on CSRF Protection. It use the good old method.
"""
import secrets
from typing import TYPE_CHECKING, Any, Callable, Coroutine

from fastapi import Request

if TYPE_CHECKING:
    from fastlife.configurator.registry import Registry  # coverage: ignore


class CSRFAttack(Exception):
    """
    An exception raised if the cookie and the csrf token hidden field did not match.
    """


def create_csrf_token() -> str:
    """A helper that create a csrf token."""
    return secrets.token_urlsafe(5)


def check_csrf(registry: "Registry") -> Callable[[Request], Coroutine[Any, Any, bool]]:
    """
    A global application dependency, that is always active.

    If you don't want csrf token, its simple don't use the
    application/x-www-form-urlencoded on a POST method.
    """

    async def check_csrf(request: Request) -> bool:
        if (
            request.method != "POST"
            or request.headers.get("content-type")
            != "application/x-www-form-urlencoded"
        ):
            return True

        cookie = request.cookies.get(registry.settings.csrf_token_name)
        if not cookie:
            raise CSRFAttack("CSRF token did not match")

        form_data = await request.form()
        value = form_data.get(registry.settings.csrf_token_name)
        if value != cookie:
            raise CSRFAttack("CSRF token did not match")

        return True

    return check_csrf



import abc
from typing import Any, Mapping, Optional, Type

from fastapi import Request
from markupsafe import Markup
from pydantic.fields import FieldInfo

from fastlife.request.model_result import ModelResult


class AbstractTemplateRenderer(abc.ABC):
    """
    An object that will be initialized by an AbstractTemplateRendererFactory,
    passing the request to process.
    """

    route_prefix: str
    """Used to buid pydantic form"""

    @abc.abstractmethod
    def render_template(
        self,
        template: str,
        *,
        globals: Optional[Mapping[str, Any]] = None,
        **params: Any,
    ) -> str:
        """
        Render the given template with the given params.

        While rendering templates, the globals parameter is keps by the instantiated
        renderer and sent to every rendering made by the request.
        This is used by the pydantic form method that will render other templates
        for the request.
        In traditional frameworks, only one template is rendered containing the whole
        pages. But, while rendering a pydantic form, every field is rendered in its
        distinct template. The template renderer keep the globals and git it back
        to every templates. This can be used to fillout options in a select without
        performing an ajax request for example.

        :param template: name of the template to render
        :param globals: some variable that will be passed to all rendered templates.
        :param params: paramaters that are limited to the main rendered templates.
        :return: The template rendering result.
        """

    @abc.abstractmethod
    def pydantic_form(
        self, model: ModelResult[Any], *, token: Optional[str] = None
    ) -> Markup:
        """
        Render an http form from a given model.

        Because template post may be give back to users with errors,
        the model is wrap in an object containing initial model, or validated model
        and a set of errors.

        this function is used inside the template directly. And it will not render the
        <form> tag so the action/httpx post is not handled byu the method..
        Somethinging like this

        ::

            <form action="" method="post">
                {{ pydantic_form(model) }}
            </form>


        :param model: model to render
        :param token: a random string that can be passed for testing purpose.
        """
        ...

    @abc.abstractmethod
    def pydantic_form_field(
        self,
        model: Type[Any],
        *,
        name: str | None,
        token: str | None,
        removable: bool,
        field: FieldInfo | None,
    ) -> Markup:
        """
        Render a field of a model inside a pydantic_form.

        Models that contains field of type Union, for instance may have
        many types of children and the form have a user interaction to choose
        which sub type to select. When the user choose the type, the sub model is
        rendered using this helper under the hood.
        """


class AbstractTemplateRendererFactory(abc.ABC):
    """
    The template render factory.
    """

    @abc.abstractmethod
    def __call__(self, request: Request) -> AbstractTemplateRenderer:
        """
        While processing an HTTP Request, a renderer object is created giving
        isolated context per request.

        :param Request: the HTTP Request to process.
        :return: The renderer object that will process that request.
        """



from typing import Any, Callable

from fastapi import Depends, Request, Response

from fastlife.configurator.registry import Registry
from fastlife.security.csrf import create_csrf_token

Template = Callable[..., Response]
TemplateEngine = Callable[["Registry", Request], Template]


def get_template(template: str, *, content_type: str = "text/html") -> TemplateEngine:
    def render_template(
        reg: "Registry",
        request: Request,
        *,
        _create_csrf_token: Callable[..., str] = create_csrf_token,
    ) -> Template:
        def parametrizer(**kwargs: Any) -> Response:
            request.scope[reg.settings.csrf_token_name] = (
                request.cookies.get(reg.settings.csrf_token_name)
                or _create_csrf_token()
            )
            data = reg.renderer(request).render_template(template, **kwargs)
            resp = Response(data, headers={"Content-Type": content_type})
            resp.set_cookie(
                reg.settings.csrf_token_name,
                request.scope[reg.settings.csrf_token_name],
                secure=request.url.scheme == "https",
                samesite="strict",
                max_age=60 * 15,
            )
            return resp

        return parametrizer

    return render_template


def template(template_path: str) -> Template:
    """
    Return a FastAPI dependency template engine ready to render the template.


    :param template_path: path to template to render by the engine setup in the regitry.
    :return: A callable accepting kwargs to pass as the context, returning a string.
    """
    return Depends(get_template(template_path))



from pydantic import BaseModel


def space_join(*segments: str) -> str:
    return " ".join(segments)


class Constants(BaseModel):
    A_CLASS: str = space_join(
        "text-primary-600",
        "hover:text-primary-500",
        "hover:underline",
        "dark:text-primary-300",
        "dark:hover:text-primary-400",
    )

    BUTTON_CLASS: str = space_join(
        "bg-primary-600",
        "px-5",
        "py-2.5",
        "font-semibold",
        "rounded-lg",
        "text-center",
        "text-sm",
        "text-white",
        "hover:bg-primary-700",
        "hover:bg-primary-200",
        "focus:outline-none",
        "focus:ring-4",
        "focus:ring-primary-300",
        "dark:bg-primary-600",
        "dark:focus:ring-primary-800",
        "dark:hover:bg-primary-700",
    )

    DETAILS_CLASS: str = "border border-neutral-100 p-4 rounded-m"

    SECONDARY_BUTTON_CLASS: str = space_join(
        "bg-neutral-300",
        "px-5",
        "py-2.5",
        "rounded-lg",
        "font-semibold",
        "text-center",
        "text-neutral-900",
        "text-sm",
        "focus:outline-none",
        "focus:ring-4",
        "focus:ring-primary-300",
        "hover:bg-neutral-200",
        "dark:bg-neutral-400",
        "dark:focus:ring-neutral-100",
        "dark:hover:bg-neutral-300",
    )

    CHECKBOX_CLASS: str = space_join(
        "bg-neutral-100",
        "border-neutral-300",
        "h-4",
        "rounded",
        "text-primary-600",
        "w-4",
        "dark:bg-neutral-700",
        "dark:border-neutral-600",
        "dark:focus:ring-primary-600",
        "dark:ring-offset-neutral-800",
        "focus:ring-2",
        "focus:ring-primary-500",
    )
    FORM_CLASS: str = "space-y-4 md:space-y-6"
    H1_CLASS: str = space_join(
        "block",
        "font-bold",
        "font-sans",
        "leading-tight",
        "pb-4",
        "text-5xl",
        "text-neutral-900",
        "tracking-tight",
        "dark:text-white",
        "md:text-4xl",
    )
    H2_CLASS: str = space_join(
        "block",
        "font-bold",
        "font-sans",
        "leading-tight",
        "pb-4",
        "text-4xl",
        "text-neutral-900",
        "tracking-tight",
        "dark:text-white",
        "md:text-4xl",
    )
    H3_CLASS: str = space_join(
        "block",
        "font-bold",
        "font-sans",
        "leading-tight",
        "pb-4",
        "text-3xl",
        "text-neutral-900",
        "tracking-tight",
        "dark:text-white",
        "md:text-3xl",
    )
    H3_SUMMARY_CLASS: str = space_join(
        "block",
        "font-bold",
        "font-sans",
        "leading-tight",
        "text-3xl",
        "text-neutral-900",
        "tracking-tight",
        "dark:text-white",
        "md:text-3xl",
    )

    H4_CLASS: str = space_join(
        "block",
        "font-bold",
        "font-sans",
        "leading-tight",
        "pb-4",
        "text-2xl",
        "text-neutral-900",
        "tracking-tight",
        "dark:text-white",
        "md:text-2xl",
    )
    H5_CLASS: str = space_join(
        "block",
        "font-bold",
        "font-sans",
        "leading-tight",
        "pb-4",
        "text-xl",
        "text-neutral-900",
        "tracking-tight",
        "dark:text-white",
        "md:text-xl",
    )
    H6_CLASS: str = space_join(
        "block",
        "font-bold",
        "font-sans",
        "leading-tight",
        "pb-4",
        "text-l",
        "text-neutral-900",
        "tracking-tight",
        "dark:text-white",
        "md:text-l",
    )
    INPUT_CLASS: str = space_join(
        "bg-neutral-50",
        "block",
        "border",
        "border-neutral-300",
        "p-2.5",
        "rounded-lg",
        "text-base",
        "text-neutral-900",
        "w-full",
        "dark:bg-neutral-700",
        "dark:border-neutral-600",
        "dark:focus:border-primary-500",
        "dark:focus:ring-primary-500",
        "dark:placeholder-neutral-400",
        "dark:text-white",
        "focus:border-primary-500",
        "focus:ring-primary-500",
    )
    LABEL_CLASS: str = space_join(
        "block",
        "font-bold",
        "mb-2",
        "text-base",
        "text-neutral-900",
        "dark:text-white",
    )
    P_CLASS: str = space_join(
        "text-base",
        "text-neutral-900",
        "dark:text-white",
    )
    RADIO_DIV_CLASS: str = "flex items-center mb-4"
    RADIO_INPUT_CLASS: str = space_join(
        "bg-neutral-100",
        "border-neutral-300",
        "w-4",
        "h-4",
        "text-primary-600",
        "focus:ring-2",
        "focus:ring-primary-500",
        "dark:bg-neutral-700",
        "dark:border-neutral-600",
        "dark:focus:ring-primary-600",
        "dark:ring-offset-neutral-800",
    )
    RADIO_LABEL_CLASS: str = (
        "ms-2 text-sm font-medium text-neutral-900 dark:text-neutral-300"
    )
    SELECT_CLASS: str = space_join(
        "bg-neutral-50",
        "block",
        "border",
        "border-neutral-300",
        "p-2.5",
        "rounded-lg",
        "text-base",
        "text-neutral-900",
        "w-full",
        "focus:border-primary-500",
        "focus:ring-primary-500",
        "dark:bg-neutral-700",
        "dark:border-neutral-600",
        "dark:focus:border-primary-500",
        "dark:focus:ring-primary-500",
        "dark:placeholder-neutral-400",
        "dark:text-white",
    )
    TABLE_CLASS: str = "table-auto w-full text-left border-colapse"
    TD_CLASS: str = "px-4 py-2 font-normal border-b dark:border-neutral-500"
    TH_CLASS: str = "px-4 py-2 font-medium border-b dark:border-neutral-500"



"""
A middleware that update the request scope for https behind a proxy.

The attempt of this middleware is to fix Starlette behavior that use client and scheme
header based on the header x-forwarded-* headers and the x-real-ip.

the x-forwarded-for header is not parsed to find the appropriate value,
the x-real-ip is used.
notethat the x-forwarded-port header is not used.

Note that uvicorn or hypercorn offer the same kind middleware.
"""

import logging
from typing import Optional, Sequence, Tuple

from starlette.types import ASGIApp, Receive, Scope, Send

from fastlife.configurator.base import AbstractMiddleware

log = logging.getLogger(__name__)


def get_header(headers: Sequence[Tuple[bytes, bytes]], key: bytes) -> Optional[str]:
    for hdr in headers:
        if hdr[0].lower() == key:
            return hdr[1].decode("latin1")
    return None


class XForwardedStar(AbstractMiddleware):
    def __init__(
        self,
        app: ASGIApp,
    ) -> None:
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] in ("http", "websocket"):
            headers = scope["headers"]
            new_vals = {
                "client": get_header(headers, b"x-real-ip"),
                "host": get_header(headers, b"x-forwarded-host"),
                "scheme": get_header(headers, b"x-forwarded-proto"),
            }
            scope.update({key: val for key, val in new_vals.items() if val is not None})

        await self.app(scope, receive, send)



from typing import Any

from fastapi import APIRouter

from fastlife.configurator.route_handler import FastlifeRoute


class FastLifeRouter(APIRouter):
    """The router used split your app in many routes."""

    def __init__(self, **kwargs: Any) -> None:
        kwargs["route_class"] = FastlifeRoute
        super().__init__(**kwargs)



from typing import Any, Sequence

from markupsafe import Markup

from fastlife.templating.renderer.abstract import AbstractTemplateRenderer

from .base import Widget


class ModelWidget(Widget[Sequence[Widget[Any]]]):
    def __init__(
        self,
        name: str,
        *,
        value: Sequence[Widget[Any]],
        error: str | None = None,
        removable: bool,
        title: str,
        token: str,
    ):
        super().__init__(
            name,
            title=title,
            value=value,
            error=error,
            removable=removable,
            token=token,
        )

    def get_template(self) -> str:
        return "pydantic_form.Model"

    def to_html(self, renderer: AbstractTemplateRenderer) -> Markup:
        """Return the html version"""
        children_widget = [child.to_html(renderer) for child in self.value or []]
        kwargs = {
            "widget": self,
            "children_widget": children_widget,
        }
        return Markup(
            renderer.render_template(self.get_template(), globals=None, **kwargs)
        )



from .abstract import AbstractTemplateRendererFactory
from .constants import Constants
from .jinjax import JinjaxTemplateRenderer

__all__ = [
    "AbstractTemplateRendererFactory",
    "Constants",
    "JinjaxTemplateRenderer",
]



"""Testing your application."""

import re
import time
from collections.abc import MutableMapping
from http.cookiejar import Cookie
from typing import Any, Iterator, Literal, Mapping, Optional, Sequence
from urllib.parse import urlencode

import bs4
import httpx
from fastapi.testclient import TestClient
from multidict import MultiDict
from starlette.types import ASGIApp

from fastlife.configurator.settings import Settings
from fastlife.middlewares.session.serializer import AbsractSessionSerializer
from fastlife.shared_utils.resolver import resolve

CookieTypes = httpx._types.CookieTypes  # type: ignore
Cookies = httpx._models.Cookies  # type: ignore


class Element:
    """Access to a dom element."""

    def __init__(self, client: "WebTestClient", tag: bs4.Tag):
        self._client = client
        self._tag = tag

    def click(self) -> "WebResponse":
        """Simulate a client to a a link. No javascript exectuted here."""
        return self._client.get(self._tag.attrs["href"])

    @property
    def node_name(self) -> str:
        """Get the node name of the dom element."""
        return self._tag.name

    @property
    def attrs(self) -> dict[str, str]:
        """Attributes of the element."""
        return self._tag.attrs

    @property
    def text(self) -> str:
        """
        Return the text of the element, with text of childs element.

        Note that the text is stripped for convenience but inner text may contains
        many spaces not manipulated here.
        """
        return self._tag.text.strip()

    @property
    def h1(self) -> "Element":
        """
        Return the h1 child element.

        Should be used on the html body element directly.
        """
        nodes = self.by_node_name("h1")
        assert len(nodes) == 1, f"Should have 1 <h1>, got {len(nodes)} in {self}"
        return nodes[0]

    @property
    def h2(self) -> Sequence["Element"]:
        """
        Return the h2 elements.
        """
        return self.by_node_name("h2")

    @property
    def form(self) -> "Element | None":
        """Get the form element of the web page."""
        return Element(self._client, self._tag.form) if self._tag.form else None

    @property
    def hx_target(self) -> Optional[str]:
        """
        Return the hx-target of the element.

        It may be set on a parent. It also resolve special case "this" and return the id
        of the element.
        """
        el: bs4.Tag | None = self._tag
        while el:
            if "hx-target" in el.attrs:
                ret = el.attrs["hx-target"]
                if ret == "this":
                    ret = el.attrs["id"]
                return ret
            el = el.parent
        return None

    def by_text(self, text: str, *, node_name: str | None = None) -> "Element | None":
        """Find the first element that match the text."""
        nodes = self.iter_all_by_text(text, node_name=node_name)
        return next(nodes, None)

    def iter_all_by_text(
        self, text: str, *, node_name: str | None = None
    ) -> "Iterator[Element]":
        """Return an iterator of all elements that match the text."""
        nodes = self._tag.find_all(string=re.compile(rf"\s*{text}\s*"))
        for node in nodes:
            if isinstance(node, bs4.NavigableString):
                node = node.parent

            if node_name:
                while node is not None:
                    if node.name == node_name:
                        yield Element(self._client, node)
                    node = node.parent
            elif node:
                yield Element(self._client, node)
        return None

    def get_all_by_text(
        self, text: str, *, node_name: str | None = None
    ) -> "Sequence[Element]":
        """Return the list of all elements that match the text."""
        nodes = self.iter_all_by_text(text, node_name=node_name)
        return list(nodes)

    def by_label_text(self, text: str) -> "Element | None":
        """Return the element which is the target of the label having the given text."""
        label = self.by_text(text, node_name="label")
        assert label is not None
        assert label.attrs.get("for") is not None
        resp = self._tag.find(id=label.attrs["for"])
        assert not isinstance(resp, bs4.NavigableString)
        return Element(self._client, resp) if resp else None

    def by_node_name(
        self, node_name: str, *, attrs: dict[str, str] | None = None
    ) -> list["Element"]:
        """
        Return the list of elements with the given node_name.

        An optional set of attributes may given and must match if passed.
        """
        return [
            Element(self._client, e) for e in self._tag.find_all(node_name, attrs or {})
        ]

    def __repr__(self) -> str:
        return f"<{self.node_name}>"

    def __str__(self) -> str:
        return str(self._tag)


class WebForm:
    """
    Handle form.

    Form are filled out and submit with methods and try to avoid invalid
    usage, such as selecting an option that don't exists is not possible here.
    Again, no javascript is executed here, but htmx attribute `hx-post` and `hx-target`
    are read while submiting to simulate it.
    """

    def __init__(self, client: "WebTestClient", origin: str, form: Element):
        self._client = client
        self._form = form
        self._origin = origin
        self._formfields: dict[str, Element] = {}
        self._formdata: MultiDict[str] = MultiDict()
        inputs = self._form.by_node_name("input")
        for input in inputs:
            self._formfields[input.attrs["name"]] = input
            if input.attrs.get("type") == "checkbox" and "checked" not in input.attrs:
                continue
            self._formdata.add(input.attrs["name"], input.attrs.get("value", ""))

        selects = self._form.by_node_name("select")
        for select in selects:
            fieldname = select.attrs["name"]
            self._formfields[fieldname] = select
            options = select.by_node_name("option")
            if "multiple" in select.attrs:
                for option in options:
                    if "selected" in option.attrs:
                        self._formdata.add(
                            fieldname, option.attrs.get("value", option.text)
                        )
            else:
                if options:
                    self._formdata[fieldname] = options[0].attrs.get(
                        "value", options[0].text
                    )
                    for option in options:
                        if "selected" in option.attrs:
                            self._formdata[fieldname] = option.attrs.get(
                                "value", option.text
                            )
                            break

        # field textearea...

    def set(self, fieldname: str, value: str) -> Any:
        """
        Set a value to an input field.

        It works for checkbox and radio as well.
        Checkbox may contains many values.
        Options of select can't be set with this method, the select method must
        be used instead.
        """
        if fieldname not in self._formfields:
            raise ValueError(f'"{fieldname}" does not exists')
        if self._formfields[fieldname].node_name == "select":
            raise ValueError(f'"{fieldname}" is a <select>, use select() instead')

        if self._formfields[fieldname].attrs.get("type") == "checkbox":
            self._formdata.add(fieldname, value)
            return

        if self._formfields[fieldname].attrs.get("type") == "radio":
            radio = self._form.by_node_name(
                "input", attrs={"type": "radio", "value": value}
            )
            if not radio:
                raise ValueError(
                    f'radio "{fieldname}" does not contains {value} option'
                )

        self._formdata[fieldname] = value

    def unset(self, fieldname: str, value: str) -> Any:
        """Unset an element. Only works with checkbox."""
        if fieldname not in self._formfields:
            raise ValueError(f'"{fieldname}" does not exists')
        if self._formfields[fieldname].node_name != "input":
            raise ValueError(f'"{fieldname}" is not a checkbox')
        if self._formfields[fieldname].attrs.get("type") != "checkbox":
            raise ValueError(f'"{fieldname}" is not a checkbox')
        values = self._formdata.popall(fieldname)
        if value not in values:
            raise ValueError(f'"{value}" not in "{fieldname}"')
        for val in values:
            if val != value:
                self._formdata[fieldname] = val

    def select(self, fieldname: str, value: str) -> Any:
        """
        Select an option, if multiple, value is added, otherwise, value is replaced.
        """
        if fieldname not in self._formfields:
            raise ValueError(f'"{fieldname}" does not exists')
        field = self._formfields[fieldname]
        if field.node_name != "select":
            raise ValueError(f"{fieldname} is a {repr(field)}, " "use set() instead")

        for option in field.by_node_name("option"):
            if option.text == value.strip():
                if "multiple" in field.attrs:
                    self._formdata.add(fieldname, value)
                else:
                    self._formdata[fieldname] = option.attrs.get("value", option.text)
                break
        else:
            raise ValueError(f'No option {value} in <select name="{fieldname}">')

    def unselect(self, fieldname: str, value: str) -> Any:
        """
        Unselect an option if multiple, otherwise an exception is raised.
        """
        if fieldname not in self._formfields:
            raise ValueError(f'"{fieldname}" does not exists')
        field = self._formfields[fieldname]

        if field.node_name != "select":
            raise ValueError(
                f"{fieldname} is a {repr(self._formfields[fieldname])}, "
                "use unset() for checkbox instead"
            )
        if "multiple" not in field.attrs:
            raise ValueError("only <select multiple> support unselect")

        for option in self._formfields[fieldname].by_node_name("option"):
            if option.text == value.strip():
                values = self._formdata.popall(fieldname)
                if value not in values:
                    raise ValueError(f'"{value}" not selected in "{fieldname}"')
                for val in values:
                    if val != value:
                        self._formdata[fieldname] = val
                break
        else:
            raise ValueError(f'No option {value} in <select name="{fieldname}">')

    def button(self, text: str, position: int = 0) -> "WebForm":
        """
        Simmulate a click on a button using the text of the button,

        and eventually a position. The button return the form and the submit()
        should be called directly.

        This is used in order to inject the value of the button in the form, usually
        done while many actions are available on a form.

        ::

            form.button("Go").submit()

        """
        buttons = self._form.get_all_by_text(text, node_name="button")
        if position >= len(buttons):
            pos = ""
            if position > 0:
                pos = f" at position {position}"
            raise ValueError(f'Button "{text}" not found{pos}')
        button = buttons[position]
        if "name" in button.attrs:
            self._formdata[button.attrs["name"]] = button.attrs.get("value", "")
        return self

    def submit(self, follow_redirects: bool = True) -> "WebResponse":
        """
        Submit the form as it has been previously filled out.
        """
        headers: dict[str, str] = {}
        target = (
            self._form.attrs.get("hx-post")
            or self._form.attrs.get("post")
            or self._origin
        )
        if "hx-post" in self._form.attrs:
            if hx_target := self._form.hx_target:
                headers["HX-Target"] = hx_target

        return self._client.post(
            target,
            data=self._formdata,
            headers=headers,
            follow_redirects=follow_redirects,
        )

    def __contains__(self, key: str) -> bool:
        """Test if a field exists in the form."""
        return key in self._formdata

    def __repr__(self) -> str:
        return repr(self._formdata)


class WebResponse:
    """Represent an http response made by the WebTestClient browser."""

    def __init__(self, client: "WebTestClient", origin: str, response: httpx.Response):
        self._client = client
        self._response = response
        self._origin = origin
        self._html: bs4.BeautifulSoup | None = None
        self._form: WebForm | None = None

    @property
    def status_code(self) -> int:
        """Http status code."""
        return self._response.status_code

    @property
    def is_redirect(self) -> bool:
        """True for any kind of http redirect status."""
        return 300 <= self._response.status_code < 400

    @property
    def content_type(self) -> str:
        """Get the content type of the response, from the header."""
        return self._response.headers.get("content-type", "").split(";").pop(0)

    @property
    def headers(self) -> httpx.Headers:
        """All http headers of the response."""
        return self._response.headers

    @property
    def text(self) -> str:
        """Http response body."""
        return self._response.text

    @property
    def html(self) -> Element:
        """Http response body as an Element."""
        if self._html is None:
            self._html = bs4.BeautifulSoup(self._response.text, "html.parser")
        return Element(self._client, self._html)

    @property
    def html_body(self) -> Element:
        """The body element of the html response."""
        body = self.html.by_node_name("body")
        assert len(body) == 1
        return body[0]

    @property
    def form(self) -> WebForm:
        """The form element of the html response."""
        if self._form is None:
            form = self.html.form
            assert form is not None
            self._form = WebForm(self._client, self._origin, form)
        return self._form

    def by_text(self, text: str, *, node_name: str | None = None) -> Element | None:
        """Search a dom element by its text."""
        return self.html.by_text(text, node_name=node_name)

    def by_label_text(self, text: str) -> Element | None:
        """Search a dom element by its associated label text."""
        return self.html.by_label_text(text)

    def by_node_name(
        self, node_name: str, *, attrs: dict[str, str] | None = None
    ) -> list[Element]:
        """List dom element having the given node name, and eventually attributes."""
        return self.html.by_node_name(node_name, attrs=attrs)


class Session(dict[str, Any]):
    """Manipulate the session of the WebTestClient browser."""

    def __init__(self, client: "WebTestClient"):
        self.client = client
        self.srlz = client.session_serializer
        self.settings = self.client.settings
        data: Mapping[str, Any]
        cookie_name = self.settings.session_cookie_name
        self.has_session = cookie_name in self.client.cookies
        if self.has_session:
            data, error = self.srlz.deserialize(
                self.client.cookies[cookie_name].encode("utf-8")
            )
            if error:
                self.has_session = False
        else:
            data = {}
        super().__init__(data)

    def __setitem__(self, __key: Any, __value: Any) -> None:
        """Initialize a value in the session of the client in order to test."""
        super().__setitem__(__key, __value)
        settings = self.settings
        data = self.serialize()
        self.client.cookies.jar.set_cookie(
            Cookie(
                version=0,
                name=settings.session_cookie_name,
                value=data,
                port=None,
                port_specified=False,
                domain=f".{settings.session_cookie_domain}",
                domain_specified=True,
                domain_initial_dot=True,
                path="/",
                path_specified=True,
                secure=False,
                expires=int(time.time() + settings.session_duration.total_seconds()),
                discard=False,
                comment=None,
                comment_url=None,
                rest={"HttpOnly": None, "SameSite": "lax"},  # type: ignore
                rfc2109=False,
            )
        )
        # this does not work
        # self.client.cookies.set(
        #     settings.session_cookie_name,
        #     data,
        #     settings.session_cookie_domain,
        #     settings.session_cookie_path,
        # )

    def serialize(self) -> str:
        return self.srlz.serialize(self).decode("utf-8")


class WebTestClient:
    """The fake browser used for testing purpose."""

    def __init__(
        self,
        app: ASGIApp,
        *,
        settings: Settings | None = None,
        cookies: CookieTypes | None = None,
    ) -> None:
        self.app = app
        if settings is None:
            settings = Settings()
            settings.domain_name = settings.domain_name or "testserver.local"
        self.testclient = TestClient(
            app, base_url=f"http://{settings.domain_name}", cookies=cookies or {}
        )
        self.settings = settings
        self.session_serializer: AbsractSessionSerializer = resolve(
            settings.session_serializer
        )(
            settings.session_secret_key,
            int(settings.session_duration.total_seconds()),
        )

    @property
    def cookies(self) -> Cookies:
        """HTTP Cookies"""
        return self.testclient.cookies

    @property
    def session(self) -> MutableMapping[str, Any]:
        """Session shared between the server and the client."""
        return Session(self)

    def request(
        self,
        method: Literal["GET", "POST", "DELETE"],
        url: str,
        *,
        content: str | None = None,
        headers: Mapping[str, str] | None = None,
        max_redirects: int = 0,
    ) -> WebResponse:
        """Perform http requests."""
        rawresp = self.testclient.request(
            method=method,
            url=url,
            headers=headers,
            content=content,
            follow_redirects=False,
        )
        resp = WebResponse(
            self,
            url,
            rawresp,
        )
        if resp.is_redirect and max_redirects > 0:
            if resp.status_code != 307:
                method = "GET"
                headers = None
                content = None
            return self.request(
                method=method,
                url=resp.headers["location"],
                content=content,
                headers=headers,
                max_redirects=max_redirects - 1,
            )
        if "HX-Redirect" in resp.headers and max_redirects > 0:
            return self.request(
                method="GET",
                url=resp.headers["HX-Redirect"],
                content=None,
                headers=headers,
                max_redirects=max_redirects - 1,
            )

        return resp

    def get(self, url: str, follow_redirects: bool = True) -> WebResponse:
        """Perform http GET request."""
        return self.request(
            "GET",
            url,
            max_redirects=int(follow_redirects) * 10,
        )

    def delete(self, url: str, follow_redirects: bool = True) -> WebResponse:
        """Perform http DELETE request."""
        return self.request(
            "DELETE",
            url,
            max_redirects=int(follow_redirects) * 10,
        )

    def post(
        self,
        url: str,
        data: MultiDict[str],
        *,
        headers: Mapping[str, Any] | None = None,
        follow_redirects: bool = True,
    ) -> WebResponse:
        """Perform http POST request in "application/x-www-form-urlencoded" format."""
        if headers is None:
            headers = {}
        return self.request(
            "POST",
            url,
            content=urlencode(data),
            headers={"Content-Type": "application/x-www-form-urlencoded", **headers},
            max_redirects=int(follow_redirects) * 10,
        )



from fastlife import Configurator, configure

from .x_forwarded import XForwardedStar


@configure
def includeme(config: Configurator) -> None:
    settings = config.registry.settings
    if settings.decode_reverse_proxy_headers:
        config.add_middleware(XForwardedStar)
